import numpy as np
import pytest
from matplotlib.colors import LinearSegmentedColormap, ListedColormap

from tastymap.utils import cmap_to_array, get_cmap, replace_match, subset_cmap


class TestGetmap:
    def test_valid(self):
        cmap = get_cmap("viridis")
        assert isinstance(cmap, ListedColormap)

    def test_invalid(self):
        with pytest.raises(ValueError, match="Unknown colormap 'invalid_cmap'."):
            get_cmap("invalid_cmap")

    def test_suggestion(self):
        with pytest.raises(
            ValueError,
            match="Did you mean one of these",
        ):
            get_cmap("virid")

    def test_mixed_case(self):
        cmap = get_cmap("ViRiDiS")
        assert isinstance(cmap, ListedColormap)

    def test_upper_case(self):
        cmap = get_cmap("VIRIDIS")
        assert isinstance(cmap, ListedColormap)

    def test_empty_string(self):
        with pytest.raises(ValueError, match="Unknown colormap ''."):
            get_cmap("")

    def test_whitespace(self):
        with pytest.raises(ValueError, match="Unknown colormap ' viridis '."):
            get_cmap(" viridis ")

    def test_non_string_input(self):
        with pytest.raises(AttributeError):
            get_cmap(123)


class TestSubsetmap:
    @pytest.fixture
    def basic_cmap(self):
        return LinearSegmentedColormap.from_list(
            "basic", ["red", "green", "blue", "yellow"]
        )

    def test_subset_with_integer(self, basic_cmap):
        subset = subset_cmap(basic_cmap, 1)
        assert len(cmap_to_array(subset)) == 2
        assert subset.name == "basic_i1"

    def test_subset_with_float(self, basic_cmap):
        subset = subset_cmap(basic_cmap, 1.5)
        assert len(cmap_to_array(subset)) == 2
        assert subset.name == "basic_i1.5"

    def test_subset_with_slice(self, basic_cmap):
        subset = subset_cmap(basic_cmap, slice(1, 3))
        assert len(cmap_to_array(subset)) == 2
        assert subset.name == "basic_i1:3:None"

    def test_subset_with_slice_step_only(self, basic_cmap):
        subset = subset_cmap(basic_cmap, slice(None, None, 2))
        assert len(cmap_to_array(subset)) == 128
        assert subset.name == "basic_i::2"

    def test_subset_with_iterable(self, basic_cmap):
        subset = subset_cmap(basic_cmap, [0, 2])
        assert len(cmap_to_array(subset)) == 2
        assert subset.name == "basic_i0,2"

    def test_subset_with_single_iterable(self, basic_cmap):
        subset = subset_cmap(basic_cmap, [2])
        assert len(cmap_to_array(subset)) == 2
        assert subset.name == "basic_i2,2"

    def test_custom_name(self, basic_cmap):
        subset = subset_cmap(basic_cmap, 1, "custom_name")
        assert subset.name == "custom_name_i1"

    def test_invalid_indices(self, basic_cmap):
        with pytest.raises(IndexError):
            subset_cmap(basic_cmap, 1000)

    def test_invalid_iterable(self, basic_cmap):
        with pytest.raises(IndexError):
            subset_cmap(basic_cmap, [0, 1000])


class TestmapToArray:
    def test_from_str(self):
        arr = cmap_to_array("viridis")
        assert isinstance(arr, np.ndarray)

    def test_from_linear_segmented_colormap(self):
        cmap = LinearSegmentedColormap.from_list("testing", ["red", "green", "blue"])
        arr = cmap_to_array(cmap)
        assert isinstance(arr, np.ndarray)

    def test_from_listed_colormap(self):
        cmap = ListedColormap(["red", "green", "blue"])
        arr = cmap_to_array(cmap)
        assert isinstance(arr, np.ndarray)

    def test_from_iterable(self):
        arr = cmap_to_array(["red", "green", "blue"])
        assert isinstance(arr, np.ndarray)


class TestReplaceMatch:
    def test_single_match(self):
        new_string, match = replace_match(r"\d+", "hello123world", "number")
        assert new_string == "helloworld"
        assert match == "123"

    def test_no_match(self):
        new_string, match = replace_match(r"\d+", "helloworld", "number")
        assert new_string == "helloworld"
        assert match == ""

    def test_multiple_matches(self):
        with pytest.raises(
            ValueError, match="Should only contain one 'number' but found .*?"
        ):
            replace_match(r"\d+", "hello123world456", "number")

    def test_empty_string(self):
        new_string, match = replace_match(r"\d+", "", "number")
        assert new_string == ""
        assert match == ""

    def test_special_regex_chars(self):
        new_string, match = replace_match(r"\.\*", "hello.*world", "regex chars")
        assert new_string == "helloworld"
        assert match == ".*"

    def test_pattern_at_start(self):
        new_string, match = replace_match(r"^\d+", "123helloworld", "number")
        assert new_string == "helloworld"
        assert match == "123"

    def test_pattern_at_end(self):
        new_string, match = replace_match(r"\d+$", "helloworld123", "number")
        assert new_string == "helloworld"
        assert match == "123"

    def test_non_string_pattern(self):
        with pytest.raises(TypeError):
            replace_match(123, "helloworld", "number")

    def test_non_string_input(self):
        with pytest.raises(TypeError):
            replace_match(r"\d+", 123, "number")



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap

from .models import ColorModel, HoloViewsTastyBar, MatplotlibTastyBar, TastyMap


def cook_tmap(
    colors_or_cmap: Sequence | str | Colormap,
    num_colors: int | None = None,
    reverse: bool = False,
    name: str | None = None,
    hue: float | None = None,
    saturation: float | None = None,
    value: float | None = None,
    bad: str | None = None,
    under: str | None = None,
    over: str | None = None,
    from_color_model: ColorModel | str | None = None,
) -> TastyMap:
    """Cook a completely new colormap or modify an existing one.

    Args:
        colors_or_cmap: A list of colors or a colormap instance or string.
        num_colors: Number of colors in the colormap. Defaults to None.
        reverse: Whether to reverse the colormap. Defaults to False.
        name: Name of the colormap; if provided, registers the cmap. Defaults to None.
        hue: Hue factor (-255 to 255) to tweak by.
        saturation: Saturation factor (-10 to 10) to tweak by.
        value: Brightness value factor (0, 3) to tweak by.
        bad: Color for bad values. Defaults to None.
        under: Color for underflow values. Defaults to None.
        over: Color for overflow values. Defaults to None.
        from_color_model: Color model of the input colormap; required if
            colors_or_cmap is an Sequence and not hexcodes. Defaults to None.

    Returns:
        TastyMap: A new TastyMap instance with the new colormap.
    """
    if isinstance(colors_or_cmap, str):
        tmap = TastyMap.from_str(colors_or_cmap)
    elif isinstance(colors_or_cmap, LinearSegmentedColormap):
        tmap = TastyMap(colors_or_cmap)
    elif isinstance(colors_or_cmap, ListedColormap):
        tmap = TastyMap.from_listed_colormap(colors_or_cmap)
    elif isinstance(colors_or_cmap, (Sequence, np.ndarray)):
        if not isinstance(colors_or_cmap[0], str) and from_color_model is None:
            raise ValueError(
                "Please specify from_color_model to differentiate "
                "between RGB and HSV color models."
            )
        if len(colors_or_cmap) == 1:
            colors_or_cmap = list(colors_or_cmap) * 2
        tmap = TastyMap.from_list(
            colors_or_cmap, color_model=from_color_model or ColorModel.RGB
        )
    else:
        raise TypeError(
            f"Expected str, LinearSegmentedColormap, ListedColormap, or Sequence; "
            f"received {type(colors_or_cmap)!r}."
        )

    if bad or under or over:
        tmap = tmap.set_extremes(bad=bad, under=under, over=over)

    if hue or saturation or value:
        tmap = tmap.tweak_hsv(hue=hue, saturation=saturation, value=value)

    if num_colors:
        tmap = tmap.resize(num_colors)

    if reverse:
        tmap = tmap.reverse()

    if name:
        tmap.cmap.name = name
        tmap.register()

    return tmap


def pair_tbar(
    plot: Any,
    colors_or_cmap_or_tmap: (Sequence | str | Colormap | TastyMap),
    bounds: slice | Sequence[float],
    labels: list[str] | None = None,
    uniform_spacing: bool = True,
    **tbar_kwargs: dict[str, Any],
):
    """Create a custom colorbar for a plot.

    Args:
        plot: A matplotlib or HoloViews plot.
        colors_or_cmap_or_tmap: A list of colors or a colormap instance or string.
        bounds: Bounds of the colorbar.
        labels: Labels for the colorbar.
        uniform_spacing: Whether to space the colors uniformly.
        tbar_kwargs: Keyword arguments for the colorbar.

    Returns:
        The plot.
    """
    tmap = colors_or_cmap_or_tmap
    if not isinstance(tmap, TastyMap):
        tmap = cook_tmap(colors_or_cmap_or_tmap)  # type: ignore

    if hasattr(plot, "axes"):
        tbar = MatplotlibTastyBar(
            tmap,
            bounds=bounds,
            labels=labels,
            uniform_spacing=uniform_spacing,
            **tbar_kwargs,  # type: ignore
        )
    elif hasattr(plot, "opts"):
        tbar = HoloViewsTastyBar(  # type: ignore
            tmap,
            bounds=bounds,
            labels=labels,
            uniform_spacing=uniform_spacing,
            **tbar_kwargs,  # type: ignore
        )
    else:
        raise NotImplementedError
    return tbar.add_to(plot)



from __future__ import annotations

from collections.abc import Sequence
from difflib import get_close_matches
from re import IGNORECASE, findall, sub

import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap
from matplotlib.pyplot import colormaps
from matplotlib.pyplot import get_cmap as _get_cmap


def get_registered_cmaps() -> dict[str, str]:
    """
    Get a mapping of registered colormaps.

    Returns:
        A mapping of registered colormaps.
    """
    return {cmap.lower(): cmap for cmap in colormaps()}


def get_cmap(cmap: str) -> Colormap:
    """
    Get a colormap by name.

    Args:
        cmap: The name of the colormap.

    Returns:
        A colormap.
    """
    lower_colormaps = get_registered_cmaps()
    try:
        return _get_cmap(lower_colormaps[cmap.lower()])
    except KeyError:
        matches = get_close_matches(cmap, lower_colormaps.values(), n=5, cutoff=0.1)
        if matches:
            raise ValueError(
                f"Unknown colormap '{cmap}'. Did you mean one of these: {matches}?"
            )
        else:
            raise ValueError(f"Unknown colormap '{cmap}'.")


def subset_cmap(
    cmap: Colormap,
    indices: int | float | slice | Sequence,
    name: str | None = None,
) -> LinearSegmentedColormap:
    """
    Subset a colormap.

    Args:
        cmap: A colormap.
        indices: The indices to subset.
        name: The name of the new colormap.

    Returns:
        A new colormap.
    """
    cmap_array = cmap_to_array(cmap)
    name = name or cmap.name
    if isinstance(indices, (int, float)):
        cmap_indices = np.array([indices] * 2).astype(int)
        name += f"_i{indices}"
    elif isinstance(indices, Sequence):
        cmap_indices = np.array(indices)
        if len(cmap_indices) == 1:
            cmap_indices = np.array([cmap_indices] * 2).astype(int)
        name += f"_i{','.join(str(i) for i in cmap_indices.flat)}"
    elif isinstance(indices, slice):
        cmap_indices = indices  # type: ignore
        step = indices.step
        start = indices.start
        stop = indices.stop
        if not indices.start and not indices.stop:
            if step == -1:
                name, r_match = replace_match(r"_r+(?=_|$)", name, "_r")
                if not r_match:
                    name += "_r"
            else:
                name += f"_i::{step}"
        else:
            name += f"_i{start}:{stop}:{step}"

    cmap_array = cmap_array[cmap_indices]
    cmap = LinearSegmentedColormap.from_list(name, cmap_array, N=len(cmap_array))
    return cmap


def cmap_to_array(
    cmap: Colormap | Sequence,
) -> np.ndarray:
    """
    Convert a colormap to an array of colors as RGB.

    Args:
        cmap: A colormap.

    Returns:
        An array of colors.
    """
    if isinstance(cmap, str):
        cmap = get_cmap(cmap)  # type: ignore

    if isinstance(cmap, LinearSegmentedColormap):
        cmap_array = cmap(np.linspace(0, 1, cmap.N))
    elif isinstance(cmap, ListedColormap):
        cmap_array = np.array(cmap.colors)
    else:
        cmap_array = np.array(cmap)
    return cmap_array


def replace_match(pattern: str, string: str, key: str) -> tuple[str, str]:
    """
    Find a pattern in a string and remove it.

    Args:
        pattern: The pattern to find.
        string: The string to search.
        key: The name of the pattern.

    Returns:
        The new string and the match.
    """
    matches = findall(pattern, string, IGNORECASE)
    if len(matches) > 1:
        raise ValueError(f"Should only contain one {key!r} but found {matches}")
    elif len(matches) == 1:
        string = sub(pattern, "", string, count=1, flags=IGNORECASE)
    return string, matches[0] if matches else ""



import ast
from io import StringIO

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import hsv_to_rgb, rgb2hex

try:
    import panel as pn  # type: ignore[import]
    import param  # type: ignore[import]
    import requests  # type: ignore[import]
    import xarray as xr  # type: ignore[import]
except ImportError:
    raise ImportError(
        "TastyKitchen additionally requires panel, param, requests, and xarray; "
        "run `pip install 'tastymap[ui]'` to install."
    )

from .core import cook_tmap, pair_tbar
from .models import ColorModel, TastyMap
from .utils import get_cmap, get_registered_cmaps

pn.extension("jsoneditor", notifications=True)
pn.Column.sizing_mode = "stretch_width"


class TastyKitchen(pn.viewable.Viewer):
    reverse = param.Boolean(
        default=False,
        doc="Whether to reverse the colormap. Defaults to False.",
        label="Reverse Colormap",
    )

    cmap = param.ObjectSelector(
        default="accent",
        doc="The colormap to use. Defaults to 'viridis'.",
    )

    colors = param.List(
        default=None,
        doc="List of colors to use for the colormap. Defaults to None.",
        item_type=(str, tuple),
    )

    from_color_model = param.ObjectSelector(
        default=ColorModel.RGB.value.upper(),
        objects=["RGB", "HSV"],
        doc="Color model of the input colors to determine whether to use RGB or HSV.",
    )

    num_colors = param.Integer(default=11, bounds=(2, 256), doc="Number of colors.")

    name = param.String(
        default=None, doc="Name of the custom colormap. Defaults to None."
    )

    bad = param.String(
        default=None,
        doc="Color for bad values. Defaults to None.",
        label="Bad Color",
    )

    under = param.String(
        default=None,
        doc="Color for underflow values. Defaults to None.",
        label="Underflow Color",
    )

    over = param.String(
        default=None,
        doc="Color for overflow values. Defaults to None.",
        label="Overflow Color",
    )

    hue = param.Number(
        default=0,
        bounds=(-255, 255),
        doc="Hue factor (-255 to 255) to tweak by.",
    )

    saturation = param.Number(
        default=1,
        bounds=(-10, 10),
        doc="Saturation factor (-10 to 10) to tweak by.",
    )

    value = param.Number(
        default=1,
        bounds=(0, 3),
        doc="Brightness value factor (0, 3) to tweak by.",
    )

    custom_name = param.String(
        default="custom_tastymap",
        doc="Name of the custom colormap. Defaults to 'custom_tastymap'.",
        label="Colormap name",
    )

    uniform_spacing = param.Boolean(
        default=False,
        doc="Whether to use uniform spacing.",
        label="Uniform Spacing Between Ticks",
    )

    bounds = param.List(
        default=None, item_type=(float, int), doc="Bounds for the colormap."
    )

    labels = param.List(default=None, item_type=str, doc="Labels for the colormap.")

    package = param.ObjectSelector(
        default="matplotlib",
        objects=["matplotlib"],
        doc="Which package to use for plotting.",
    )

    _tmap = param.ClassSelector(class_=TastyMap, doc="The tastymap.", precedence=-1)

    _registered_cmaps = param.Dict(doc="Registered colormaps.", precedence=-1)

    _active_index = param.Integer(default=0, precedence=-1)

    def __init__(self, **params):
        self._plot = pn.pane.Matplotlib(tight=True, sizing_mode="stretch_width")
        self._reference_image = pn.pane.Image(height=300, sizing_mode="stretch_width")
        self._palette_box = pn.FlexBox(min_height=100, sizing_mode="stretch_width")
        self._tmap_html = pn.pane.HTML(height=115)
        self._mpl_code_md = pn.pane.Markdown(
            name="Matplotlib",
            min_height=300,
            sizing_mode="stretch_width",
            margin=(-20, 0, 0, 0),
        )
        self._tmap_code_md = pn.pane.Markdown(
            name="TastyMap",
            min_height=300,
            sizing_mode="stretch_width",
            margin=(-20, 0, 0, 0),
        )
        self._history_box = pn.FlexBox(height=100)
        super().__init__(**params)

        # cmap widgets

        cmaps = {cmap_name: get_cmap(cmap_name) for cmap_name in get_registered_cmaps()}
        cmaps = self._sort_cmaps(cmaps)
        self.cmap_input = pn.widgets.ColorMap(
            options=cmaps,
            ncols=2,
            swatch_width=55,
            name="Colormap",
            margin=(5, 20, 5, 20),
            sizing_mode="stretch_width",
        )
        cmap_button = pn.widgets.Button(
            name="Import Colormap Palette",
            margin=(5, 20, 5, 20),
            sizing_mode="stretch_width",
        )
        self.cmap_method = pn.widgets.RadioButtonGroup(
            options=[
                "Overwrite",
                "Prepend",
                "Append",
            ],
            margin=(5, 20, 5, 20),
            sizing_mode="stretch_width",
        )
        cmap_widgets = pn.Column(self.cmap_input, self.cmap_method, cmap_button)

        self.cmap_input.param.watch(self._update_cmap, "value")
        cmap_button.on_click(self._use_cmap_palette)

        # colors widgets
        self.colors_select = pn.widgets.JSONEditor.from_param(
            self.param.colors,
            sizing_mode="stretch_width",
            margin=(5, 30, 5, 20),
        )
        from_color_model_select = pn.widgets.Select.from_param(
            self.param.from_color_model,
            name="Color Model",
            margin=(5, 5, 5, 20),
        )
        colors_input = pn.widgets.TextAreaInput(
            placeholder="Enter colors, separated by new lines.",
            margin=(5, 5, 5, 20),
            auto_grow=True,
            max_rows=3,
        )
        colors_picker = pn.widgets.ColorPicker(
            name="Color Picker",
            sizing_mode="stretch_width",
            margin=(5, 30, 5, 20),
        )
        colors_upload = pn.widgets.FileInput(
            accept=".rgb,.txt",
            sizing_mode="stretch_width",
            margin=(10, 30, 5, 20),
        )
        colors_clear = pn.widgets.Button(
            name="Clear",
            sizing_mode="stretch_width",
            margin=(5, 30, 5, 20),
        )
        colors_widgets = pn.Column(
            from_color_model_select,
            pn.Tabs(
                ("Text", colors_input),
                ("Pick", colors_picker),
                ("Upload", colors_upload),
                ("Clear", colors_clear),
            ),
            self.colors_select,
        )

        colors_input.param.watch(self._add_color, "value")
        colors_picker.param.watch(self._add_color, "value")
        colors_upload.param.watch(self._add_color, "value")
        colors_clear.on_click(lambda event: setattr(self.colors_select, "value", []))

        # tmap widgets

        tmap_parameters = [
            "reverse",
            "num_colors",
            "hue",
            "saturation",
            "value",
            "bad",
            "under",
            "over",
        ]
        tmap_widgets = pn.Param(
            self,
            parameters=tmap_parameters,
            show_name=False,
            widgets={
                "reverse": {
                    "type": pn.widgets.Toggle,
                    "sizing_mode": "stretch_width",
                    "margin": (5, 10, 5, 10),
                },
                "bad": {"placeholder": "black"},
                "under": {"placeholder": "blue"},
                "over": {"placeholder": "red"},
            },
        )

        # tbar widgets

        tbar_parameters = ["bounds", "labels", "uniform_spacing", "package"]
        tbar_widgets = pn.Param(
            self,
            parameters=tbar_parameters,
            show_name=False,
            widgets={
                "uniform_spacing": {
                    "type": pn.widgets.Toggle,
                    "sizing_mode": "stretch_width",
                    "margin": (5, 10, 5, 10),
                },
                "bounds": {"placeholder": "[230, 273.15, 291, 300, 330]"},
                "labels": {
                    "placeholder": "['freezing', 'cool', 'comfortable', 'warm', 'hot']"
                },
            },
        )

        # finalize widgets

        name_input = pn.widgets.TextInput.from_param(
            self.param.custom_name,
            sizing_mode="stretch_width",
            margin=(5, 30, 5, 20),
        )
        register_button = pn.widgets.Button(
            name="Register Colormap for Session",
            sizing_mode="stretch_width",
            margin=(5, 30, 5, 20),
        )
        self.colors_download = pn.widgets.FileDownload(
            filename=f"{self.custom_name}_hexcodes.txt",
            callback=pn.bind(self._download_colors, self.param._tmap),
            sizing_mode="stretch_width",
            margin=(5, 30, 5, 20),
        )
        finalize_widgets = pn.Column(
            name_input,
            register_button,
            self.colors_download,
        )

        register_button.on_click(self._register_tmap)

        # reference widgets

        file_input = pn.widgets.FileInput(
            accept=".png, .jpg, .jpeg, .gif, .bmp, .svg, .tiff, .tif",
            margin=(10, 30, 5, 20),
            sizing_mode="stretch_width",
        )
        url_input = pn.widgets.TextInput(
            placeholder="Enter a URL to an image to use as reference",
            margin=(10, 30, 5, 20),
            sizing_mode="stretch_width",
        )
        reference_tabs = pn.Tabs(("URL", url_input), ("File", file_input))

        url_input.param.watch(self._add_reference, "value")
        file_input.param.watch(self._add_reference, "value")

        # layout widgets

        self._colors_or_cmap_tabs = pn.Tabs(
            ("Colormap", cmap_widgets), ("Colors", colors_widgets), dynamic=True
        )
        self._colors_or_cmap_tabs.link(self, active="_active_index", bidirectional=True)

        self._widgets = pn.Column(
            self._colors_or_cmap_tabs,
            pn.layout.Divider(),
            tmap_widgets,
            pn.layout.Divider(),
            tbar_widgets,
            pn.layout.Divider(),
            finalize_widgets,
        )

        self.palette_tabs = pn.Tabs(
            pn.Column(
                pn.pane.HTML("<h2>🎨 Palette</h2>"), self._palette_box, name="Palette"
            ),
            pn.Column(
                pn.pane.HTML("<h2>⏱️ History</h2>"), self._history_box, name="History"
            ),
        )
        self.colorbar_col = pn.Column(
            pn.pane.HTML("<h2>🍭 Colormap</h2>"), self._tmap_html, name="Colormap"
        )
        self.image_tabs = pn.Tabs(
            pn.Column(
                pn.pane.HTML("<h2>🖼️ Example Output</h2>"),
                self._plot,
                name="Example Output",
            ),
            pn.Column(
                pn.pane.HTML("<h2>🌇 Reference Image</h2>"),
                reference_tabs,
                self._reference_image,
                name="Reference Image",
            ),
            pn.Column(
                pn.Row(
                    pn.pane.HTML("<h2>🖥️ Matplotlib Code</h2>"),
                ),
                self._mpl_code_md,
                name="Matplotlib Code",
            ),
            pn.Column(
                pn.Row(
                    pn.pane.HTML("<h2>🖥️ TastyMap Code</h2>"),
                ),
                self._tmap_code_md,
                name="TastyMap Code",
            ),
            dynamic=True,
        )
        self.param.trigger("cmap")

    # event methods

    def _update_cmap(self, event):
        if not event.new:
            return
        self.cmap = self.cmap_input.value_name

    def _use_cmap_palette(self, event):
        num_colors = min(self.num_colors, 16)
        palette = self._tmap.resize(num_colors).to_model("hex").tolist()

        value = self.colors_select.value
        if isinstance(self.colors_select.value, dict):
            value = list(value)

        if self.cmap_method.value == "Prepend":
            value = palette + value
        elif self.cmap_method.value == "Overwrite":
            value = palette[:]
        else:
            value = value + palette
        self.colors_select.param.update(value=value)
        self._active_index = 1
        self._add_to_history(palette)

    def _add_color(self, event):
        new_event = event.new
        if not new_event:
            return

        if isinstance(new_event, bytes):
            new_event = new_event.decode("utf-8")

        value = self.colors_select.value
        if isinstance(value, dict):
            value = list(value)

        if " " in new_event:
            new_event = new_event.strip()

        if "\n" in new_event:
            new_event = new_event.splitlines()

        if not isinstance(new_event, list):
            new_event = [new_event]

        processed_colors = []
        for color in new_event:
            if not color.strip() or color.startswith("#"):
                continue
            try:
                if " " in color or "," in color:
                    color = np.array(
                        ast.literal_eval(",".join(color.strip().split()))
                    ).astype(float)
                    if any(c > 1 for c in color):
                        color /= 255
                    color = tuple(color.round(2))
            except Exception as exc:
                pn.state.notifications.error(str(exc))
                continue
            processed_colors.append(color)
        try:
            self.colors_select.param.update(value=value + processed_colors)
            self._add_to_history(processed_colors)
        except ValueError as exc:
            if "invalid" in str(exc).lower():
                self.colors_select.param.update(value=value)
                pn.state.notifications.error(str(exc))
        finally:
            if isinstance(event.obj, pn.widgets.TextInput):
                event.obj.value = ""

        try:
            self.num_colors = len(self.colors_select.value)
        except Exception:
            pass

    def _add_reference(self, event):
        if not event.new:
            return

        try:
            if isinstance(event.new, str):
                response = requests.get(event.new)
                response.raise_for_status()
                self._reference_image.object = response.content
                event.obj.value = ""
            else:
                self._reference_image.object = event.new
        except Exception as exc:
            pn.state.notifications.error(str(exc))
        finally:
            self.image_tabs.active = 1

    def _add_to_history(self, value):
        new_history = self._history_box.objects + self._render_colors(value)
        self._history_box.objects = new_history[-26:]

    def _register_tmap(self, event):
        self._tmap.register(name=self.custom_name)
        options = self.cmap_input.options
        options[self.custom_name] = self._tmap.cmap
        self.cmap_input.options = self._sort_cmaps(options)
        pn.state.notifications.success(
            f"Registered {self.custom_name} for this session and it can now be "
            f"accessed under the Colormap tab.",
            10000,
        )

    def _sort_cmaps(self, cmaps):
        return dict(sorted(cmaps.items(), key=lambda item: item[0]))

    def _download_colors(self, tmap):
        colors_string = "\n".join(tmap.to_model("hex"))
        buf = StringIO(colors_string)
        return buf

    # param methods
    def _render_colors(self, colors):
        color_background_tuples = []

        for color in colors:
            prefix = ""
            background_color = color
            if isinstance(color, tuple):
                if any(c > 1 for c in color):
                    color = tuple(c / 255 for c in color)
                prefix = "RGB<br>"
                if self.from_color_model == "HSV":
                    background_color = hsv_to_rgb(color)
                    prefix = "HSV<br>"
                background_color = rgb2hex(background_color)
            color_background_tuples.append((f"{prefix}{color}", background_color))
        return [
            pn.pane.HTML(
                f"<center style='background-color: lightgrey; "
                f"color: black;'>{color}</center>",
                styles={
                    "background-color": background_color,
                    "font-size": "0.75em",
                },
                height=75,
                width=75,
                margin=(5, 5, 15, 5),
            )
            for color, background_color in color_background_tuples
        ]

    @pn.depends(
        "reverse",
        "cmap",
        "colors",
        "from_color_model",
        "num_colors",
        "hue",
        "saturation",
        "value",
        "bad",
        "under",
        "over",
        "_active_index",
        watch=True,
    )
    def _cook_tmap(self):
        if self._active_index == 1:
            colors_or_colormap = self.colors
        else:
            colors_or_colormap = get_cmap(self.cmap)

        if not colors_or_colormap:
            return

        self._tmap = cook_tmap(
            colors_or_cmap=colors_or_colormap,
            num_colors=self.num_colors,
            reverse=self.reverse,
            bad=self.bad,
            under=self.under,
            over=self.over,
            hue=self.hue,
            saturation=self.saturation,
            value=self.value,
            from_color_model=self.from_color_model,
        )

        if self._active_index == 1:
            colors = self.colors
        else:
            colors = self._tmap.resize(min(self.num_colors, 26)).to_model("hex")

        self._palette_box.objects = self._render_colors(colors)
        self._tmap_html.object = self._tmap.cmap._repr_html_()

        colors_string = "',\n    '".join(self._tmap.to_model("hex"))
        self._mpl_code_md.object = (
            f"```python\n"
            f"from matplotlib.colors import LinearSegmentedColormap\n"
            f"colors = [\n    '{colors_string}'\n]\n"
            f"cmap = LinearSegmentedColormap.from_list({self.custom_name!r}, colors, N={self.num_colors})\n"  # noqa: E501
            f"```\n"
        )
        self._tmap_code_md.object = (
            f"```python\n"
            f"from tastymap import cook_tmap\n"
            f"colors = [\n    '{colors_string}'\n]\n"
            f"tmap = cook_tmap(colors, name={self.custom_name!r}, num_colors={self.num_colors})\n"  # noqa: E501
            f"cmap = tmap.cmap\n"
            f"```\n"
        )

    @pn.depends("_tmap", "bounds", "labels", "uniform_spacing", watch=True)
    def _pair_tbar(self):
        fig, ax = plt.subplots(facecolor="whitesmoke")
        ds = xr.tutorial.open_dataset("air_temperature")["air"].isel(time=0)
        self._mappable = ds.plot(ax=ax, add_colorbar=False)
        if self.bounds is None:
            ini = ds.min().round(0)
            end = ds.max().round(0)
            if self.num_colors > 18:
                bounds = slice(ini, end)
            else:
                bounds = np.linspace(ini, end, min(self.num_colors - 1, 18)).tolist()
        else:
            bounds = self.bounds
        self._mappable = pair_tbar(
            self._mappable,
            self._tmap,
            bounds=bounds,
            labels=self.labels if self.labels else None,
            uniform_spacing=self.uniform_spacing,
        )
        self._plot.object = fig
        plt.close(fig)

    @pn.depends("custom_name", watch=True)
    def _update_filename(self):
        self.colors_download.filename = f"{self.custom_name}_hexcodes.txt"

    def __panel__(self):
        return pn.template.FastListTemplate(
            sidebar=[self._widgets],
            main=[self.palette_tabs, self.colorbar_col, self.image_tabs],
            sidebar_width=350,
            title="👨‍🍳 TastyKitchen",
            main_max_width="clamp(800px, 80vw, 1150px)",
        )

    def serve(self, port: int = 8888, show: bool = True, **kwargs):
        pn.serve(self.__panel__(), port=port, show=show, **kwargs)



import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from matplotlib.pyplot import get_cmap

from tastymap.core import cook_tmap, pair_tbar
from tastymap.models import TastyMap


class TestCookTmap:
    def test_cook_from_string(self):
        tmap = cook_tmap("viridis")
        assert isinstance(tmap, TastyMap)

    def test_cook_from_string_reversed(self):
        tmap = cook_tmap("viridis_r")
        assert isinstance(tmap, TastyMap)
        # Check if the colormap is reversed by comparing the first color
        assert cook_tmap("viridis_r")[0] == cook_tmap("viridis")[255]

    def test_cook_from_listed_colormap(self):
        cmap_input = ListedColormap(["red", "green", "blue"])
        tmap = cook_tmap(cmap_input)
        assert isinstance(tmap, TastyMap)

    def test_cook_from_linear_segmented_colormap(self):
        cmap_input = LinearSegmentedColormap.from_list(
            "testmap", ["red", "green", "blue"]
        )
        tmap = cook_tmap(cmap_input)
        assert isinstance(tmap, TastyMap)

    def test_cook_from_list(self):
        cmap_input = ["red", "green", "blue"]
        tmap = cook_tmap(cmap_input, num_colors=28)
        assert isinstance(tmap, TastyMap)
        assert len(tmap) == 28

    def test_cook_from_list_no_color_model(self):
        cmap_input = [(0.0, 1.0, 1.0), (0.5, 1.0, 1.0), (1.0, 1.0, 1.0)]
        with pytest.raises(ValueError):
            cook_tmap(cmap_input)

    def test_r_flag_with_reverse_true(self):
        tmap = cook_tmap("viridis_r", reverse=True)
        assert isinstance(tmap, TastyMap)
        assert np.all(tmap.to_model("rgba")[0] == get_cmap("viridis")(0))

    def test_r_flag_with_reverse_false(self):
        tmap = cook_tmap("viridis_r", reverse=False)
        assert isinstance(tmap, TastyMap)
        assert np.all(tmap.to_model("rgba")[0] == get_cmap("viridis")(256))

    def test_with_num_colors(self):
        tmap = cook_tmap("viridis", num_colors=20)
        assert isinstance(tmap, TastyMap)
        assert len(tmap) == 20

    def test_register(self):
        tmap = cook_tmap("viridis", name="test")
        assert plt.get_cmap("test") == tmap.cmap

    def test_non_iterable_input(self):
        with pytest.raises(TypeError):
            cook_tmap(12345)

    def test_bad_under_over(self):
        tmap = cook_tmap("viridis", under="red", over="blue", bad="green")
        tmap.cmap.get_under().tolist() == [1.0, 0.0, 0.0, 1.0]
        tmap.cmap.get_over().tolist() == [0.0, 0.0, 1.0, 1.0]
        tmap.cmap.get_bad().tolist() == [0.0, 1.0, 0.0, 1.0]


class TestPairTbar:
    def test_pair_tbar(self):
        fig, ax = plt.subplots()
        img = ax.imshow(np.random.random((10, 10)))
        tmap = cook_tmap(["red", "green", "blue"])
        pair_tbar(
            img, tmap, bounds=[0, 1], labels=["a", "b", "c"], uniform_spacing=False
        )
        assert len(fig.axes) == 2

    def test_pair_tbar_list(self):
        fig, ax = plt.subplots()
        img = ax.imshow(np.random.random((10, 10)))
        pair_tbar(
            img,
            ["red", "green", "blue"],
            bounds=[0, 1],
            labels=["a", "b", "c"],
            uniform_spacing=False,
        )
        assert len(fig.axes) == 2



import argparse

from .ui import TastyKitchen


def main():
    parser = argparse.ArgumentParser(description="Serve the TastyKitchen UI")
    parser.add_argument("command", help="The command to run", choices=["ui"])
    args = parser.parse_args()

    if args.command == "ui":
        TastyKitchen().serve(port=8888, show=True)


if __name__ == "__main__":
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Generator, Sequence
from enum import Enum
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colormaps
from matplotlib.cm import ScalarMappable
from matplotlib.colors import (
    BoundaryNorm,
    Colormap,
    LinearSegmentedColormap,
    ListedColormap,
    Normalize,
    hsv_to_rgb,
    rgb2hex,
    rgb_to_hsv,
)
from matplotlib.ticker import FuncFormatter

from .utils import cmap_to_array, get_cmap, subset_cmap


class ColorModel(Enum):
    """Enumeration for different color models.

    Attributes:
        rgba: Red, green, blue, alpha.
        rgb: Red, green, blue.
        hsv: Hue, saturation, value.
        hex: Hexcode.
    """

    RGBA = "rgba"
    RGB = "rgb"
    HSV = "hsv"
    HEX = "hex"


class TastyMap:
    """A class to represent and manipulate colormaps in a tasty manner.

    Attributes:
        cmap: The colormap object.
        cmap_array: RGB array representation of the colormap.
    """

    def __init__(
        self,
        cmap: Colormap,
        name: str | None = None,
    ):
        """Initializes a TastyMap instance.

        Args:
            cmap: The colormap to be used.
            name: Name of the colormap. Defaults to the name of the provided colormap.
        """
        if not isinstance(cmap, LinearSegmentedColormap):
            raise TypeError(
                f"Expected LinearSegmentedColormap; received {type(cmap)!r}."
            )

        cmap = cmap.copy()
        cmap.name = name or cmap.name
        self.cmap: Colormap = cmap
        self._cmap_array = cmap_to_array(cmap)

    @classmethod
    def from_str(cls, string: str) -> TastyMap:
        """Creates a TastyMap instance from a string name.

        Args:
            string: Name of the colormap.

        Returns:
            TastyMap: A new TastyMap instance.
        """
        cmap = get_cmap(string)  # type: ignore
        cmap_array = cmap_to_array(cmap)

        new_name = string
        cmap = LinearSegmentedColormap.from_list(
            new_name, cmap_array, N=len(cmap_array)
        )
        return TastyMap(cmap, name=new_name)

    @classmethod
    def from_list(
        cls,
        colors: Sequence,
        name: str = "custom_tastymap",
        color_model: ColorModel | str = ColorModel.RGBA,
    ) -> TastyMap:
        """Creates a TastyMap instance from a list of colors.

        Args:
            colors: List of colors.
            name: Name of the colormap. Defaults to "custom_tastymap".

        Returns:
            TastyMap: A new TastyMap instance.
        """
        if not isinstance(colors, (Sequence, np.ndarray)):
            raise TypeError(f"Expected Sequence; received {type(colors)!r}.")
        if colors is None or len(colors) == 0:  # type: ignore
            raise ValueError("Must provide at least one color.")
        cmap_array = cmap_to_array(colors)

        if color_model in (ColorModel.HSV, ColorModel.HSV.value):
            cmap_array = hsv_to_rgb(cmap_array)

        cmap = LinearSegmentedColormap.from_list(name, cmap_array, N=len(cmap_array))
        return TastyMap(cmap)

    @classmethod
    def from_listed_colormap(
        cls,
        listed_colormap: ListedColormap,
        name: str = "custom_tastymap",
    ) -> TastyMap:
        """Creates a TastyMap instance from a ListedColormap.

        Args:
            listed_colormap: The colormap to be used.
            name: Name of the colormap. Defaults to "custom_tastymap".

        Returns:
            TastyMap: A new TastyMap instance.
        """
        return cls.from_list(listed_colormap.colors, name=name)  # type: ignore

    def _from_list_with_extremes(self, *args, **kwargs) -> LinearSegmentedColormap:
        """Creates a TastyMap instance from a list of colors with extreme values."""
        cmap = LinearSegmentedColormap.from_list(*args, **kwargs)
        cmap.set_extremes(
            bad=self.cmap.get_bad(),  # type: ignore
            under=self.cmap.get_under(),  # type: ignore
            over=self.cmap.get_over(),  # type: ignore
        )
        return cmap

    def resize(self, num_colors: int) -> TastyMap:
        """Resizes the colormap to a specified number of colors.

        Args:
            num_colors: Number of colors to resize to.

        Returns:
            TastyMap: A new TastyMap instance with the interpolated colormap.
        """
        cmap = self._from_list_with_extremes(
            self.cmap.name, self._cmap_array, N=num_colors
        )
        return TastyMap(cmap)

    def register(self, name: str | None = None, echo: bool = True) -> TastyMap:
        """Registers the colormap with matplotlib.

        Returns:
            TastyMap: A new TastyMap instance with the registered colormap.
        """
        tmap = self.rename(name) if name else self
        colormaps.register(self.cmap, name=tmap.cmap.name, force=True)
        if echo:
            print(
                f"Successfully registered the colormap; "
                f"to use, set `cmap={tmap.cmap.name!r}` in your plot."
            )
        return tmap

    def rename(self, name: str) -> TastyMap:
        """Renames the colormap.

        Args:
            name (str): New name for the colormap.

        Returns:
            TastyMap: A new TastyMap instance with the renamed colormap.
        """
        return TastyMap(self.cmap, name=name)

    def reverse(self) -> TastyMap:
        """Reverses the colormap.

        Returns:
            TastyMap: A new TastyMap instance with the reversed colormap.
        """
        return self[::-1]

    def to_model(self, color_model: ColorModel | str) -> np.ndarray:
        """Converts the colormap to a specified color model.

        Args:
            color_model: The color model to convert to.

        Returns:
            np.ndarray: Array representation of the colormap
                in the specified color model.
        """
        if isinstance(color_model, str):
            try:
                color_model = ColorModel(color_model.lower())
            except ValueError:
                raise ValueError(
                    f"Invalid color model: {color_model!r}; "
                    f"select from: {[cm.value for cm in ColorModel]}."
                )

        if color_model == ColorModel.RGBA:
            return self._cmap_array
        elif color_model == ColorModel.RGB:
            return self._cmap_array[:, :3]
        elif color_model == ColorModel.HSV:
            return rgb_to_hsv(self._cmap_array[:, :3])
        elif color_model == ColorModel.HEX:
            return np.apply_along_axis(
                rgb2hex, 1, self._cmap_array[:, :3]  # type: ignore
            )

    def set_extremes(
        self,
        bad: str | None = None,
        under: str | None = None,
        over: str | None = None,
    ) -> TastyMap:
        """Sets the colors for bad, underflow, and overflow values.

        Args:
            bad: Color for bad values. Defaults to None.
            under: Color for underflow values. Defaults to None.
            over: Color for overflow values. Defaults to None.

        Returns:
            TastyMap: A new TastyMap instance with the updated colormap.
        """
        cmap = self.cmap.copy()
        cmap.set_extremes(bad=bad, under=under, over=over)  # type: ignore
        return TastyMap(cmap)

    def tweak_hsv(
        self,
        hue: float | None = None,
        saturation: float | None = None,
        value: float | None = None,
        name: str | None = None,
    ) -> TastyMap:
        """Tweaks the hue, saturation, and value of the colormap.

        Args:
            hue: Hue factor (-255 to 255) to tweak by.
            saturation: Saturation factor (-10 to 10) to tweak by.
            value: Brightness value factor (0, 3) to tweak by.
            name: Name of the new colormap.

        Returns:
            TastyMap: A new TastyMap instance with the tweaked colormap.
        """
        cmap_array = self._cmap_array.copy()
        cmap_array[:, :3] = rgb_to_hsv(cmap_array[:, :3])
        if hue is not None:
            hue /= 255
            if abs(hue) > 1:
                raise ValueError("Hue must be between -255 and 255 (non-inclusive).")
            cmap_array[:, 0] = (cmap_array[:, 0] + hue) % 1
        if saturation is not None:
            if abs(saturation) > 10:
                raise ValueError("Saturation must be between -10 and 10.")
            cmap_array[:, 1] = cmap_array[:, 1] * saturation
        if value is not None:
            if value < 0 or value > 3:
                raise ValueError("Value must be between 0 and 3.")
            cmap_array[:, 2] = cmap_array[:, 2] * value
        cmap_array[:, :3] = hsv_to_rgb(np.clip(cmap_array[:, :3], 0, 1))
        cmap = self._from_list_with_extremes(
            name or self.cmap.name, cmap_array, N=len(cmap_array)
        )
        return TastyMap(cmap)

    def __iter__(self) -> Generator[np.ndarray, None, None]:
        """Iterates over the colormap.

        Yields:
            np.ndarray: An array of colors.
        """
        yield from self._cmap_array

    def __getitem__(self, indices: int | float | slice | Sequence) -> TastyMap:
        """Gets a subset of the colormap.

        Args:
            indices: Indices to subset the colormap.

        Returns:
            TastyMap: A new TastyMap instance with the subset colormap.
        """
        cmap = subset_cmap(self.cmap, indices)
        return TastyMap(cmap)

    def _repr_html_(self) -> str:
        """Returns an HTML representation of the colormap.

        Returns:
            str: HTML representation of the colormap.
        """
        return self.cmap._repr_html_()

    def __add__(self, hue: float) -> TastyMap:
        """Adds a hue factor to the colormap.

        Args:
            hue: Hue factor to add.

        Returns:
            TastyMap: A new TastyMap instance with the hue
                added to the colormap.
        """
        return self.tweak_hsv(hue=hue)

    def __sub__(self, hue: float) -> TastyMap:
        """Subtracts a hue factor to the colormap.

        Args:
            hue: Hue factor to subtract.

        Returns:
            TastyMap: A new TastyMap instance with the hue
                subtracted from the colormap.
        """
        return self.tweak_hsv(hue=-hue)

    def __mul__(self, saturation: float) -> TastyMap:
        """Multiplies a saturation factor to the colormap.

        Args:
            saturation: Saturation factor to multiply.

        Returns:
            TastyMap: A new TastyMap instance with the saturation
                multiplied to the colormap.
        """
        return self.tweak_hsv(saturation=saturation)

    def __truediv__(self, saturation: float) -> TastyMap:
        """Divides a saturation factor to the colormap.

        Args:
            saturation: Saturation factor to divide.

        Returns:
            TastyMap: A new TastyMap instance with the saturation
                divided from the colormap.
        """
        return self.tweak_hsv(saturation=1 / saturation)

    def __pow__(self, value: float) -> TastyMap:
        """Raises the brightness value factor to the colormap.

        Args:
            value: Brightness value factor to raise.

        Returns:
            TastyMap: A new TastyMap instance with the brightness value
                raised to the colormap.
        """
        return self.tweak_hsv(value=value)

    def __invert__(self) -> TastyMap:
        """Reverses the colormap.

        Returns:
            TastyMap: A new TastyMap instance with the reversed colormap.
        """
        return self.reverse()

    def __and__(self, tmap: TastyMap) -> TastyMap:
        """Combines two TastyMap instances.

        Args:
            tmap: Another TastyMap instance to combine with.

        Returns:
            TastyMap: A new TastyMap instance with the combined colormap.
        """
        if not isinstance(tmap, TastyMap):
            raise TypeError(
                f"Can only combine TastyMap instances; received {type(tmap)!r}."
            )
        name = self.cmap.name + "_" + tmap.cmap.name
        cmap_array = np.concatenate([self._cmap_array, cmap_to_array(tmap.cmap)])
        cmap = self._from_list_with_extremes(name, cmap_array, N=len(cmap_array))
        return TastyMap(cmap)

    def __or__(self, num_colors: int) -> TastyMap:
        """Interpolates the colormap to a specified number of colors.

        Args:
            num_colors: Number of colors to resize to.

        Returns:
            TastyMap: A new TastyMap instance with the interpolated colormap.
        """
        return self.resize(num_colors)

    def __lshift__(self, name: str) -> TastyMap:
        """Renames the colormap.

        Args:
            name: New name for the colormap.

        Returns:
            TastyMap: A new TastyMap instance with the renamed colormap.
        """
        return self.rename(name)

    def __rshift__(self, name: str) -> TastyMap:
        """Registers the colormap with matplotlib.

        Args:
            name: Name of the colormap.

        Returns:
            TastyMap: A new TastyMap instance with the registered colormap.
        """
        return self.register(name)

    def __mod__(self, color_model: str) -> np.ndarray:
        """Converts the colormap to a specified color model.

        Args:
            color_model: The color model to convert to.

        Returns:
            np.ndarray: Array representation of the colormap
                in the specified color model.
        """
        return self.to_model(color_model)

    def __len__(self) -> int:
        """Returns the number of colors in the colormap.

        Returns:
            int: Number of colors in the colormap.
        """
        return len(self._cmap_array)

    def __eq__(self, other: Any) -> bool:
        """Checks if two TastyMap instances are equal.

        Args:
            other: Another TastyMap instance to compare with.

        Returns:
            bool: True if the two TastyMap instances are equal; False otherwise.
        """
        if not isinstance(other, TastyMap):
            return False
        cmap_array = other._cmap_array
        return bool(np.all(self._cmap_array == cmap_array))

    def __str__(self) -> str:
        """Returns the name of the colormap.

        Returns:
            str: Name of the colormap.
        """
        return f"{self.cmap.name} ({len(self)} colors)"

    def __repr__(self) -> str:
        """Returns a string representation of the TastyMap instance.

        Returns:
            str: String representation of the TastyMap instance.
        """
        return f"TastyMap({self.cmap.name!r})"


class TastyBar(ABC):
    def __init__(
        self,
        tmap: TastyMap,
        bounds: slice | Sequence[float],
        labels: list[str] | None = None,
        uniform_spacing: bool = True,
    ):
        self.tmap = tmap
        self.bounds = bounds
        self.labels = labels
        self.uniform_spacing = uniform_spacing

    @abstractmethod
    def add_to(self, plot: Any):
        """Adds a colorbar to a plot.

        Args:
            plot: A plot.
        """


class MatplotlibTastyBar(TastyBar):
    def __init__(
        self,
        tmap: TastyMap,
        bounds: slice | Sequence[float],
        labels: list[str] | None = None,
        uniform_spacing: bool = True,
        center: bool | None = None,
        extend: Literal["both", "neither", "min", "max"] = "both",
        clip: bool | None = None,
        **colorbar_kwargs: dict[str, Any],
    ):
        """Initializes a MatplotlibTastyBar instance.

        Args:
            tmap: A TastyMap instance.
            bounds: Bounds for the colorbar.
            labels: Labels for the colorbar. Defaults to None.
            uniform_spacing: Whether to use uniform spacing for the colorbar.
                Defaults to True.
            center: Whether to center the colorbar. Defaults to None.
            extend: Whether to extend the colorbar. Defaults to "both".
            clip: Whether to clip the colorbar. Defaults to None.
            **colorbar_kwargs: Keyword arguments for the colorbar.
        """
        super().__init__(tmap, bounds, labels, uniform_spacing)
        self.extend = extend
        self.clip = clip
        self.colorbar_kwargs = colorbar_kwargs
        self.spacing = "uniform" if uniform_spacing else "proportional"

        num_colors = len(self.tmap)
        is_slice = isinstance(self.bounds, slice)
        if is_slice:
            vmin = self.bounds.start  # type: ignore
            vmax = self.bounds.stop  # type: ignore
            step = self.bounds.step  #  type: ignore
            if step is None:
                num_ticks = min(num_colors - 1, 11)
                ticks = np.linspace(vmin, vmax, num_ticks)
            else:
                ticks = np.arange(vmin, vmax + step, step)
        else:
            ticks = np.array(self.bounds)
            ticks.sort()
            vmin, vmax = ticks[0], ticks[-1]

        if center is None and not is_slice:
            center = False

        if clip is None:
            clip = self.extend == "neither"

        norm = None
        if center is None:
            norm = Normalize(vmin=vmin, vmax=vmax, clip=clip)
            if is_slice:
                ticks = None  # let matplotlib decide
        elif center:
            norm_bins = ticks + 0.5
            norm_bins = np.insert(norm_bins, 0, norm_bins[0] - 1)
            norm = BoundaryNorm(norm_bins, num_colors, clip=clip, extend=self.extend)
            if labels is None:
                labels = ticks.copy()
            ticks = norm_bins[:-1] + np.diff(norm_bins) / 2
        else:
            norm = BoundaryNorm(ticks, num_colors, clip=clip, extend=self.extend)

        format = None
        if labels is not None:
            format = FuncFormatter(
                lambda _, index: labels[index]  # type: ignore
                if index < len(labels)  # type: ignore
                else ""
            )

        self.norm = norm
        self.ticks = ticks
        self.format = format

    @property
    def plot_settings(self):
        """Keyword arguments for the plot."""
        return dict(cmap=self.tmap.cmap, norm=self.norm)

    @property
    def colorbar_settings(self):
        """Keyword arguments for the colorbar."""
        colorbar_kwargs = dict(
            cmap=self.tmap.cmap,
            norm=self.norm,
            ticks=self.ticks,
            format=self.format,
            spacing=self.spacing,
        )
        colorbar_kwargs.update(self.colorbar_kwargs)
        return colorbar_kwargs

    def add_to(self, plot: ScalarMappable) -> ScalarMappable:
        """Adds a colorbar to a plot.

        Args:
            plot: A matplotlib ax.
        """
        plot_settings = self.plot_settings
        plot.cmap = plot_settings["cmap"]
        plot.norm = plot_settings["norm"]
        plt.colorbar(plot, **self.colorbar_settings)
        return plot


class HoloViewsTastyBar(TastyBar):
    def __init__(
        self,
        tmap: TastyMap,
        bounds: slice | Sequence[float],
        labels: list[str] | None = None,
        uniform_spacing: bool = True,
    ):
        """Initializes a HoloViewsTastyBar instance.

        Args:
            tmap: A TastyMap instance.
            bounds: Bounds for the colorbar.
            labels: Labels for the colorbar. Defaults to None.
            uniform_spacing: Whether to use uniform spacing for the colorbar.
                Defaults to True.
        """
        super().__init__(tmap, bounds, labels, uniform_spacing)

        from bokeh import models  # type: ignore

        self._models = models
        self.palette = self.tmap.to_model("hex").tolist()

        self.factors = None
        self.major_label_overrides = None

        num_colors = len(self.tmap)
        is_slice = isinstance(self.bounds, slice)
        if is_slice:
            vmin = self.bounds.start  # type: ignore
            vmax = self.bounds.stop  # type: ignore
            step = self.bounds.step  #  type: ignore
            if step is None:
                num_ticks = min(num_colors - 1, 11)
                ticks = np.linspace(vmin, vmax, num_ticks)
            else:
                ticks = np.arange(vmin, vmax + step, step)
        else:
            ticks = np.array(self.bounds)
            num_ticks = len(ticks)
        self.ticks = ticks.tolist()

        num_labels = len(labels) if labels else 0
        if uniform_spacing:
            if labels is None:
                self.factors = [
                    f"{self.ticks[i]} - {self.ticks[i + 1]}"
                    for i in range(len(self.ticks) - 1)
                ]
            else:
                self.factors = labels
                if not num_labels == num_ticks - 1:
                    raise ValueError(
                        f"Number of labels must be one less than the number of ticks; "
                        f"received {num_labels} labels and {num_ticks} ticks."
                    )
        elif not uniform_spacing and labels:
            if not num_labels == num_ticks:
                raise ValueError(
                    f"Number of labels must be equal to the number of ticks; "
                    f"received {num_labels} labels and {num_ticks} ticks."
                )
            self.major_label_overrides = dict(zip(self.ticks, labels))

    def _hook(self, hv_plot, _):
        plot = hv_plot.handles["plot"]
        mapper = self._models.CategoricalColorMapper(
            palette=self.palette,
            factors=self.factors,
        )
        color_bar = self._models.ColorBar(color_mapper=mapper)
        plot.right[0] = color_bar

    @property
    def opts_settings(self):
        """Keyword arguments for opts."""
        opts_kwargs = dict(
            cmap=self.palette,
            color_levels=self.ticks,
            clim=(self.ticks[0], self.ticks[-1]),
            colorbar=True,
        )
        colorbar_opts = dict(ticker=self._models.FixedTicker(ticks=self.ticks))
        if self.uniform_spacing:
            opts_kwargs["hooks"] = [self._hook]
        elif self.major_label_overrides:
            colorbar_opts["major_label_overrides"] = self.major_label_overrides

        if colorbar_opts:
            opts_kwargs["colorbar_opts"] = colorbar_opts
        return opts_kwargs

    def add_to(self, plot):
        """Adds a colorbar to a plot.

        Args:
            plot: A HoloViews plot.
        """
        return plot.opts(**self.opts_settings)



"""Color palettes for your palate"""

from .core import cook_tmap, pair_tbar
from .models import TastyBar, TastyMap
from .ui import TastyKitchen

__version__ = "0.2.0"

__all__ = ["cook_tmap", "pair_tbar", "TastyMap", "TastyBar", "TastyKitchen"]



import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.colors import BoundaryNorm, LinearSegmentedColormap, Normalize
from matplotlib.ticker import FuncFormatter

from tastymap.models import ColorModel, MatplotlibTastyBar, TastyMap


@pytest.fixture
def tmap():
    cmap = LinearSegmentedColormap.from_list("testmap", ["red", "green", "blue"])
    return TastyMap(cmap)


class TestTastyMap:
    def test_init(self, tmap):
        assert tmap.cmap.name == "testmap"
        assert len(tmap._cmap_array) == 256

    def test_from_str(self):
        tmap = TastyMap.from_str("viridis")
        assert tmap.cmap.name == "viridis"

    def test_from_list(self):
        colors = ["red", "green", "blue"]
        tmap = TastyMap.from_list(colors)
        assert tmap.cmap.name == "custom_tastymap"
        assert len(tmap._cmap_array) == 3

    def test_from_list_hsv(self):
        colors = [(0.0, 1.0, 1.0), (0.5, 1.0, 1.0), (1.0, 1.0, 1.0)]
        tmap = TastyMap.from_list(colors, color_model="hsv")
        assert tmap.cmap.name == "custom_tastymap"
        assert len(tmap._cmap_array) == 3

    def test_empty_colormap(self):
        with pytest.raises(ValueError):
            TastyMap.from_list([])

    def test_invalid_color(self):
        with pytest.raises(ValueError):
            TastyMap.from_list(["red", "not_a_color"])

    def test_invalid_colormap_name(self):
        with pytest.raises(ValueError):
            TastyMap.from_str("not_a_real_colormap")

    def test_non_iterable_colors(self):
        with pytest.raises(TypeError):
            TastyMap.from_list(123)

    def test_non_linearsegmented_colormap(self):
        with pytest.raises(TypeError):
            TastyMap("not_a_colormap")

    def test_interpolate(self, tmap):
        interpolated = tmap.resize(10)
        assert len(interpolated._cmap_array) == 10

    def test_reverse(self, tmap):
        reversed_map = tmap.reverse()
        assert reversed_map._cmap_array[0].tolist() == tmap._cmap_array[-1].tolist()

    def test_to(self, tmap):
        rgba_array = tmap.to_model(ColorModel.RGBA)
        assert rgba_array.shape == (256, 4)
        rgb_array = tmap.to_model(ColorModel.RGB)
        assert rgb_array.shape == (256, 3)
        hsv_array = tmap.to_model(ColorModel.HSV)
        assert hsv_array.shape == (256, 3)
        hex_array = tmap.to_model(ColorModel.HEX)
        assert hex_array.shape == (256,)

    def test_set_bad(self, tmap):
        tmap = tmap.set_extremes(bad="black", under="black", over="black")
        assert tmap.cmap.get_bad().tolist() == [0.0, 0.0, 0.0, 1.0]
        assert tmap.cmap.get_over().tolist() == [0.0, 0.0, 0.0, 1.0]
        assert tmap.cmap.get_under().tolist() == [0.0, 0.0, 0.0, 1.0]

    def test_getitem(self, tmap):
        subset = tmap[10:20]
        assert len(subset._cmap_array) == 10

    def test_and(self, tmap):
        cmap2 = LinearSegmentedColormap.from_list("testmap2", ["yellow", "cyan"])
        tmap2 = TastyMap(cmap2)
        combined = tmap & tmap2
        assert len(combined._cmap_array) == 256 + 256

    def test_len(self, tmap):
        assert len(tmap) == 256

    def test_str(self, tmap):
        assert str(tmap) == "testmap (256 colors)"

    def test_repr(self, tmap):
        assert repr(tmap) == "TastyMap('testmap')"

    def test_add_non_tastymap(self, tmap):
        with pytest.raises(TypeError):
            tmap + "not_a_tastymap"

    def test_unsupported_color_model(self, tmap):
        with pytest.raises(ValueError):
            tmap.to_model("unsupported_model")

    def test_invalid_indices(self, tmap):
        with pytest.raises(IndexError):
            tmap[1000:2000]
        with pytest.raises(TypeError):
            tmap["not_an_index"]

    def test_invalid_color_model(self, tmap):
        with pytest.raises(ValueError):
            tmap.to_model("not_a_real_color_model")

    def test_tweak_hue(self, tmap):
        tweaked = tmap.tweak_hsv(hue=50)
        assert isinstance(tweaked, TastyMap)

    def test_tweak_saturation(self, tmap):
        tweaked = tmap.tweak_hsv(saturation=5)
        assert isinstance(tweaked, TastyMap)

    def test_tweak_value(self, tmap):
        tweaked = tmap.tweak_hsv(value=2)
        assert isinstance(tweaked, TastyMap)

    def test_tweak_all(self, tmap):
        tweaked = tmap.tweak_hsv(hue=50, saturation=5, value=2)
        assert isinstance(tweaked, TastyMap)

    def test_tweak_edge_values(self, tmap):
        tweaked_min_hue = tmap.tweak_hsv(hue=-255)
        tweaked_max_hue = tmap.tweak_hsv(hue=255)
        tweaked_min_saturation = tmap.tweak_hsv(saturation=-10)
        tweaked_max_saturation = tmap.tweak_hsv(saturation=10)
        tweaked_min_value = tmap.tweak_hsv(value=0)
        tweaked_max_value = tmap.tweak_hsv(value=3)

        assert isinstance(tweaked_min_hue, TastyMap)
        assert isinstance(tweaked_max_hue, TastyMap)
        assert isinstance(tweaked_min_saturation, TastyMap)
        assert isinstance(tweaked_max_saturation, TastyMap)
        assert isinstance(tweaked_min_value, TastyMap)
        assert isinstance(tweaked_max_value, TastyMap)

    def test_tweak_out_of_range(self, tmap):
        with pytest.raises(ValueError):
            tmap.tweak_hsv(hue=300)
        with pytest.raises(ValueError):
            tmap.tweak_hsv(saturation=20)
        with pytest.raises(ValueError):
            tmap.tweak_hsv(value=4)

    def test_empty_string(self):
        with pytest.raises(ValueError):
            TastyMap.from_str("")

    def test_non_string_non_colormap(self):
        with pytest.raises(TypeError):
            TastyMap(12345)

    def test_arithmetic_with_non_numeric(self, tmap):
        with pytest.raises(TypeError):
            tmap + "string"
        with pytest.raises(TypeError):
            tmap - "string"
        with pytest.raises(TypeError):
            tmap * "string"
        with pytest.raises(TypeError):
            tmap / "string"

    def test_from_list_empty_colors(self):
        with pytest.raises(ValueError, match="Must provide at least one color."):
            TastyMap.from_list([])

    def test_and_operator_with_non_tastymap(self):
        tmap = TastyMap.from_str("viridis")
        with pytest.raises(TypeError):
            tmap & "some_string"

    def test_invert(self):
        tmap = TastyMap.from_str("viridis")
        inverted = ~tmap
        np.testing.assert_equal(inverted._cmap_array, tmap._cmap_array[::-1])

    def test_pow(self):
        tmap = TastyMap.from_str("viridis")
        result = tmap**2
        assert result == tmap.tweak_hsv(value=2)

    def test_eq_operator_with_non_tastymap(self):
        tmap = TastyMap.from_str("viridis")
        assert not (tmap == "some_string")

    def test_or_operator(self):
        tmap = TastyMap.from_str("viridis")
        result = tmap | 10
        assert len(result) == 10

    def test_lshift_operator(self):
        tmap = TastyMap.from_str("viridis")
        result = tmap << "new_name"
        assert result.cmap.name == "new_name"

    def test_rshift_operator(self):
        tmap = TastyMap.from_str("viridis")
        result = tmap >> "new_name"
        assert result.cmap.name == "new_name"

    def test_mod_operator(self):
        tmap = TastyMap.from_str("viridis")
        result = tmap % "rgb"
        assert result.shape[1] == 3

    def test_len_tmap(self):
        tmap = TastyMap.from_str("viridis")
        assert len(tmap) == 256

    def test_str_tmap(self):
        tmap = TastyMap.from_str("viridis")
        assert str(tmap) == "viridis (256 colors)"

    def test_repr_tmap(self):
        tmap = TastyMap.from_str("viridis")
        assert repr(tmap) == "TastyMap('viridis')"

    def test_iter(self):
        tmap = TastyMap.from_str("viridis")
        for color in tmap:
            assert isinstance(color, np.ndarray)

    def test_resize_with_extremes(self):
        tmap = TastyMap.from_str("viridis")
        result = tmap.set_extremes(bad="black", under="black", over="black").resize(10)
        assert result.cmap.get_bad().tolist() == [0.0, 0.0, 0.0, 1.0]
        assert result.cmap.get_over().tolist() == [0.0, 0.0, 0.0, 1.0]
        assert result.cmap.get_under().tolist() == [0.0, 0.0, 0.0, 1.0]


class TestMatplotlibTastyBar:
    @pytest.fixture
    def tmap(self):
        cmap = LinearSegmentedColormap.from_list("testmap", ["red", "green", "blue"])
        return TastyMap(cmap)

    def test_init_provided_ticks(self, tmap):
        tmap_bar = MatplotlibTastyBar(tmap, bounds=[0, 4, 18])
        np.testing.assert_equal(tmap_bar.ticks, [0, 4, 18])
        assert isinstance(tmap_bar.norm, BoundaryNorm)
        assert tmap_bar.format is None
        assert not tmap_bar.norm.clip
        assert tmap_bar.norm.extend == "both"

    def test_init_provided_ticks_and_center(self, tmap):
        tmap_bar = MatplotlibTastyBar(tmap, bounds=[0, 4, 18], center=True)
        np.testing.assert_equal(tmap_bar.ticks, [0, 2.5, 11.5])
        assert isinstance(tmap_bar.norm, BoundaryNorm)
        assert isinstance(tmap_bar.format, FuncFormatter)
        assert not tmap_bar.norm.clip
        assert tmap_bar.norm.extend == "both"

    def test_init_not_provided_ticks(self, tmap):
        tmap_bar = MatplotlibTastyBar(tmap, bounds=slice(0, 18, 4))
        assert tmap_bar.ticks is None
        assert isinstance(tmap_bar.norm, Normalize)
        assert tmap_bar.format is None
        assert not tmap_bar.norm.clip
        assert tmap_bar.norm.vmin == 0
        assert tmap_bar.norm.vmax == 18

    def test_init_not_provided_ticks_no_step(self, tmap):
        tmap_bar = MatplotlibTastyBar(tmap, bounds=slice(0, 18))
        assert tmap_bar.ticks is None
        assert isinstance(tmap_bar.norm, Normalize)
        assert tmap_bar.format is None
        assert not tmap_bar.norm.clip
        assert tmap_bar.norm.vmin == 0
        assert tmap_bar.norm.vmax == 18

    def test_plot_settings(self, tmap):
        tmap_bar = MatplotlibTastyBar(tmap, bounds=[0, 4, 18])
        plot_settings = tmap_bar.plot_settings
        assert len(plot_settings) == 2
        assert plot_settings["cmap"] == tmap.cmap
        assert plot_settings["norm"] == tmap_bar.norm

    def test_colorbar_settings(self, tmap):
        tmap_bar = MatplotlibTastyBar(tmap, bounds=[0, 4, 18])
        colorbar_settings = tmap_bar.colorbar_settings
        assert len(colorbar_settings) == 5
        np.testing.assert_equal(colorbar_settings["ticks"], tmap_bar.ticks)
        assert colorbar_settings["format"] == tmap_bar.format
        assert colorbar_settings["norm"] == tmap_bar.norm
        assert colorbar_settings["spacing"] == "uniform"
        assert isinstance(colorbar_settings["norm"], BoundaryNorm)

    def test_add_to(self, tmap):
        fig, ax = plt.subplots()
        img = ax.imshow(np.random.rand(10, 10))
        tmap_bar = MatplotlibTastyBar(tmap, bounds=[0, 4, 18])
        tmap_bar.add_to(img)
        assert len(fig.axes) == 2
